/* 
   Include files needed for the Prometheus preconditioner:
     pcimpl.h - private include file intended for use by all preconditioners 
*/

/*
  build in petsc BOPT=O/g with: make libfast shared BOPT=O/g CC="mpCC_r -qrtti=dyna -+"

*/ 
#include "src/ksp/pc/pcimpl.h"   /*I "petscpc.h" I*/
#include "Prometheus_LinSysCore.h"
#include "prometheus.hh"

PETSC_EXTERN_CXX_BEGIN

/* -------------------------------------------------------------------------- */
/*
   PCCreate_Prometheus_private - create Prometheus object
   
   Input Parameter:
.  pc - the preconditioner context
*/
PetscErrorCode PCCreate_Prometheus_private( PC pc )
{
  PetscErrorCode ierr;

  PetscFunctionBegin;

  /*
    Creates the private data structure for this preconditioner and
    attach it to the PC object.
  */
  assert(pc->data==NULL);

  Prometheus_LinSysCore *lsc = new Prometheus_LinSysCore( pc->comm );
  pc->data  = (void*)lsc;

  // will create prometheus (if necc.)
  char *argv[] = {"-ksp_type preonly"};
  ierr = lsc->parameters( 0, argv ); CHKERRQ(ierr);
  lsc->mg_levels_ = 10; // hard wires for 'mg'

  /*
    Logs the memory usage; this is not needed but allows PETSc to 
    monitor how much memory is being used for various purposes.
  */
  PetscLogObjectMemory(pc,sizeof(Prometheus));

  PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCSetUp_Prometheus - Prepares for the use of the Prometheus preconditioner
   by setting data structures and options.   
   
   Input Parameter:
.  pc - the preconditioner context

   Application Interface Routine: PCSetUp()

   Notes:
   The interface routine PCSetUp() is not usually called directly by
   the user, but instead is called by PCApply() if necessary.
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetUp_Prometheus"
PetscErrorCode PCSetUp_Prometheus( PC pc )
{
  Prometheus_LinSysCore *lsc = (Prometheus_LinSysCore*)pc->data; 
  PetscErrorCode ierr; PetscInt bs, my0, tt, ii, npe = lsc->Comm_->np();
  
  PetscFunctionBegin;

  // create base
  assert(lsc!=NULL);
  if (pc->setupcalled == 0) { // allocate space the first time this is ever called
    lsc->prom_->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_PROM]);
    pc->setupcalled = 1;
    // 
    Mat mat = pc->mat; assert(mat != NULL); // graph
    ierr = MatGetBlockSize( mat, &bs );               CHKERRQ( ierr );
    ierr = MatGetOwnershipRange( mat, &my0, &tt ); CHKERRQ(ierr);
    const PetscInt nloceq = tt - my0, nloc = nloceq/bs; assert(nloceq%bs==0);
    {// construct LSC
      PromMap map( nloc, bs, *lsc->Comm_ );
      const int *proc_gnode = map.proc_gnode();
      int proc_block[PROM_BSZ], proc_geq[PROM_BSZ];
      for( ii = proc_block[0] = proc_geq[0] = 0 ; ii < npe ; ii++ ){
	proc_block[ii+1] = proc_gnode[ii+1];   // + CR[ii+1]
	proc_geq[ii+1] = bs*proc_gnode[ii+1]; // + CR[ii+1]
      }
      ierr = lsc->setGlobalOffsets(npe+1, proc_gnode, proc_geq, proc_block);
      CHKERRQ(ierr);
    }

    //
    tt = 1;
    ierr = lsc->setNumRHSVectors( 1, &tt );         CHKERRQ(ierr);
    {  // construct graph
      int **blkColIndices,*blkRowLengths,*ptRowsPerBlkRow,xx,gid,*pb;
      const PetscInt *cols; PetscInt ncols;
      ierr = PetscMalloc( (nloc+1)*sizeof(int*), &blkColIndices ); CHKERRQ(ierr);
      ierr = PetscMalloc( (nloc+1)*sizeof(int), &blkRowLengths ); CHKERRQ(ierr);
      ierr = PetscMalloc( (nloc+1)*sizeof(int), &ptRowsPerBlkRow ); CHKERRQ(ierr);
      for( xx = 0, gid = my0 ; xx < nloc ; xx++, gid += bs ) {
	ierr = MatGetRow( mat, gid, &ncols, &cols, NULL ); CHKERRQ(ierr);
	tt = blkRowLengths[xx] = ncols/bs;
	ptRowsPerBlkRow[xx] = bs;
	ierr = PetscMalloc( (tt+1)*sizeof(int), &pb ); CHKERRQ(ierr);
	blkColIndices[xx] = pb;
	for( ii = 0 ; ii < ncols ; ii += bs ) *pb++ = cols[ii]/bs;
	MatRestoreRow( mat, gid, &ncols, &cols, NULL ); CHKERRQ(ierr );
      }
      // call FEI setMatrixStructure

      ierr = lsc->setMatrixStructure( NULL, NULL, blkColIndices, 
				      blkRowLengths, ptRowsPerBlkRow, mat );
      CHKERRQ(ierr);
      
      // set coordinates
      if( lsc->coords_ != NULL ) {
	ierr = lsc->putNodalFieldData( -3, 3, NULL, nloc, lsc->coords_ ); 
	CHKERRQ(ierr);
	PetscFree( lsc->coords_ );lsc->coords_ = NULL;
      }

      // clean up
      for( xx = 0 ; xx < nloc ; xx++ ) {
	ierr = PetscFree(blkColIndices[xx]);  CHKERRQ(ierr);
      }
      ierr = PetscFree( blkColIndices );  CHKERRQ(ierr);
      ierr = PetscFree( blkRowLengths );  CHKERRQ(ierr);
      ierr = PetscFree( ptRowsPerBlkRow );  CHKERRQ(ierr);
    }
    lsc->prom_->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_PROM]);
  }

  // factor
  if ( pc->setupcalled == 1 ) { // 'factor' matrix 
    pc->setupcalled = 2;
    // factor 
    ierr = lsc->prom_->grid0_->stiffness_->SetDirty(); CHKERRQ(ierr);
    ierr = lsc->matrixLoadComplete(); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}
/* -------------------------------------------------------------------------- */
/*
   PCSetCoordinates_Prometheus - 
   
   Input Parameter:
   .  pc - the preconditioner context
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetCoordinates_Prometheus"
PetscErrorCode PCSetCoordinates_Prometheus( PC pc, PetscReal *coords )
{
  Prometheus_LinSysCore *lsc = (Prometheus_LinSysCore*)pc->data; 
  PetscErrorCode ierr;  PetscInt bs, my0, tt;

  PetscFunctionBegin;  

  Mat mat = pc->mat; assert(mat != NULL); // graph
  ierr = MatGetBlockSize( mat, &bs );               CHKERRQ( ierr );
  ierr = MatGetOwnershipRange( mat, &my0, &tt ); CHKERRQ(ierr);
  const PetscInt nloceq = tt - my0, nloc = nloceq/bs; assert(nloceq%bs == 0);

  // put coordinates
  //ierr = lsc->putNodalFieldData(-3, 3, NULL, nloc, coords ); CHKERRQ(ierr);
  ierr = PetscMalloc( (nloceq+1)*sizeof(double), &lsc->coords_ ); CHKERRQ(ierr);
  for(tt=0;tt<nloceq;tt++)lsc->coords_[tt]=coords[tt];
  
  PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCSetFromOptions_Prometheus - Set options from options database - not used!
   Prometheus automatically reads parameters from the database

   Input Parameter:
.  pc - the preconditioner context
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetFromOptions_Prometheus"
PetscErrorCode PCSetFromOptions_Prometheus(PC pc)
{
  PetscErrorCode ierr;

  PetscFunctionBegin;  
  //ierr = PetscOptionsHead("Prometheus options"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCSetUp_Prometheus_Symmetric - Allocates the vector needed to store the
   inverse of the square root of the diagonal entries of the matrix.  This
   is used for symmetric application of the Prometheus preconditioner.

   Input Parameter:
.  pc - the preconditioner context
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetUp_Prometheus_Symmetric"
PetscErrorCode PCSetUp_Prometheus_Symmetric(PC pc)
{
  //PetscFunctionBegin;
  SETERRQ(1,"PCSetUp_Prometheus_Symmetric not supported by Prometheus");
  //PetscFunctionReturn(0);
}
/* -------------------------------------------------------------------------- */
/*
   PCSetUp_Prometheus_NonSymmetric - Allocates the vector needed to store the
   inverse of the diagonal entries of the matrix.  This is used for left of
   right application of the Prometheus preconditioner.

   Input Parameter:
.  pc - the preconditioner context
*/
#undef __FUNCT__  
#define __FUNCT__ "PCSetUp_Prometheus_NonSymmetric"
PetscErrorCode PCSetUp_Prometheus_NonSymmetric(PC pc)
{
  PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PCSetUp_Prometheus(pc);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
/* -------------------------------------------------------------------------- */
/*
   PCApply_Prometheus - Applies the Prometheus preconditioner to a vector.

   Input Parameters:
.  pc - the preconditioner context
.  x - input vector

   Output Parameter:
.  y - output vector

   Application Interface Routine: PCApply()
 */
#undef __FUNCT__  
#define __FUNCT__ "PCApply_Prometheus"
PetscErrorCode PCApply_Prometheus( PC pc, Vec x, Vec y )
{
  Prometheus_LinSysCore *lsc = (Prometheus_LinSysCore*)pc->data; 
  Prometheus *prom = lsc->prom_;
  PetscErrorCode ierr; MPI_Comm wcomm = lsc->MPIComm_;

  PetscFunctionBegin;
  lsc->nSolves_++;

  // get RHS
  PromVector *work = lsc->getRHS();
  if( prom->left_ != NULL ) { // scale f <- f * D^{-1/2}
    ierr = VecPointwiseMult( x, prom->left_->vec_ , work->vec_ ); CHKERRQ(ierr);
  }
  else {
    ierr = VecCopy( x,  work->vec_ ); CHKERRQ(ierr);
  }

  // solve
  prom->perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FEI_SOLVE]);
  PromVector *XX = lsc->getSolution(); 
  ierr = XX->Set( 0.0 );   CHKERRQ(ierr);  assert(prom->solver_ != NULL);
  ierr = prom->solver_->pc_->Apply( work, XX, TRUE ); CHKERRQ(ierr);
  prom->perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FEI_SOLVE]);

  // return solution
  if( prom->right_ != NULL ) { // scale u <- D^{-1/2} * u
    ierr = VecPointwiseMult( XX->vec_, prom->right_->vec_, y );  
    CHKERRQ(ierr);
  }
  else {
    ierr = VecCopy( XX->vec_, y );  CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCApplySymmetricLeftOrRight_Prometheus - Applies the left or right part of a
   symmetric preconditioner to a vector.

   Input Parameters:
.  pc - the preconditioner context
.  x - input vector

   Output Parameter:
.  y - output vector

   Application Interface Routines: PCApplySymmetricLeft(), PCApplySymmetricRight()
*/
#undef __FUNCT__  
#define __FUNCT__ "PCApplySymmetricLeftOrRight_Prometheus"
PetscErrorCode PCApplySymmetricLeftOrRight_Prometheus(PC pc,Vec ,Vec )
{ 
  //PetscFunctionBegin;
  SETERRQ(1,"PCSetUp_Prometheus_Symmetric not supported by Prometheus");
  //PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCDestroy_Prometheus - Destroys the private context for the Prometheus preconditioner
   that was created with PCCreate_Prometheus().

   Input Parameter:
.  pc - the preconditioner context

   Application Interface Routine: PCDestroy()
*/
#undef __FUNCT__  
#define __FUNCT__ "PCDestroy_Prometheus"
PetscErrorCode PCDestroy_Prometheus(PC pc)
{
  Prometheus_LinSysCore *lsc = (Prometheus_LinSysCore*)pc->data; 

  PetscFunctionBegin;
  /*
      Free the private data structure that was hanging off the PC
  */
  delete lsc; pc->data = NULL;
  PetscFunctionReturn(0);
}

/* -------------------------------------------------------------------------- */
/*
   PCView_Prometheus - 

   Input Parameter:
.  pc - the preconditioner context

   Application Interface Routine: PCView()
*/
#undef __FUNCT__  
#define __FUNCT__ "PCView_Prometheus"
PetscErrorCode PCView_Prometheus( PC pc, PetscViewer viewer)
{
  Prometheus_LinSysCore *lsc = (Prometheus_LinSysCore*)pc->data; 
  PetscErrorCode ierr; PetscTruth     iascii;

  PetscFunctionBegin;
  ierr = PetscTypeCompare((PetscObject)viewer,PETSC_VIEWER_ASCII,&iascii);
  CHKERRQ(ierr);
  if( iascii ) {
    if( lsc->verbose_ > 0 ) {
      ierr = lsc->prom_->solver_->PrintInfo( lsc->Comm_->np() ); CHKERRQ(ierr);
    }
  } else {
    SETERRQ1(PETSC_ERR_SUP,"Viewer type %s not supported for Prometheus",
	     ((PetscObject)viewer)->type_name);
  }
  PetscFunctionReturn(0);
}

PETSC_EXTERN_CXX_END

