//  $Id: prom_petsc.hh,v 1.47 2005/05/19 19:55:17 adams Exp $
//  Author:             Mark F. Adams
// Copyright (c) 2000 by Mark F. Adams 
//  Filename:           prom_petsc.hh
//
#ifndef __PROM_PETSC_H__
#define __PROM_PETSC_H__

#if defined(PROM_USE_C_PETSC)
extern "C" {
#endif
  
#if (PETSC_VERSION_RELEASE == 1) && (PETSC_VERSION_MAJOR == 2) && (PETSC_VERSION_MINOR == 3) && (PETSC_VERSION_SUBMINOR == 0)
#include "src/vec/impls/mpi/pvecimpl.h"
#else
#include "src/vec/vec/impls/mpi/pvecimpl.h"
#endif
  //#include "src/ksp/kspimpl.h"
#include "src/ksp/pc/impls/mg/mgimpl.h"
#include "src/ksp/ksp/impls/cg/cgctx.h"
#include "petscksp.h"
#include "petscmg.h"
#include "petscpc.h"
#include "petscmat.h"
#include "src/mat/impls/baij/mpi/mpibaij.h"
#include "src/mat/impls/sbaij/mpi/mpisbaij.h"
#include "src/mat/impls/aij/mpi/mpiaij.h"
  
#if defined(PROM_USE_C_PETSC)
}
#endif

typedef IS PromIS;

#include "prom_base.hh"

//! PromIS methods for index sets for PC blocks
#define PromISGetSize ISGetLocalSize
#define PromISGetIndices ISGetIndices
#define PromISRestoreIndices ISRestoreIndices
#define PromISCreateGeneral ISCreateGeneral
#define PromISDestroy ISDestroy

#undef __FUNCT__
#define __FUNCT__ "PromComm" 
//! PromComm (petsc)
class PromComm : public PromComm_base
{
public:
  PromComm( MPI_Comm comm ) : mpi_comm_( comm ) {
    assert(mpi_comm_!=MPI_COMM_NULL); 
  }
  MPI_Comm operator()() const { return mpi_comm_; }
  MPI_Comm MPIComm() const { return mpi_comm_; }
  int mype() const { 
    int tt; assert(mpi_comm_ != MPI_COMM_NULL);
    MPI_Comm_rank( mpi_comm_, &tt); 
    return tt; 
  }
  int np()const{ 
    assert(mpi_comm_!=MPI_COMM_NULL); int tt; 
    MPI_Comm_size(mpi_comm_,&tt); 
    return tt;
  }
  int getNewTag() const { 
    int tag; PetscCommGetNewTag( mpi_comm_, &tag ); return tag;
  }
  int getNewTag(MPI_Comm comm) const { 
    int tag; PetscCommGetNewTag( comm, &tag ); return tag;
  }
  const MPI_Comm mpi_comm_;
};

#undef __FUNCT__
#define __FUNCT__ "PromMap" 
//! PromMap (petsc)
class PromMap : public PromMap_base
{
  friend class PromMatrix;
  friend class PromComplexMatrix;
public:
  PromMap( int nloc, const int ElementSize, const PromComm &comm ) 
    : PromMap_base(nloc,ElementSize*nloc,comm), id_leq_(NULL), bsize_(ElementSize) {}
  PromMap( int nloc, const PromComm &comm, const int nloceq ) // hack for VBR
    : PromMap_base(nloc,nloceq,comm), id_leq_(NULL), bsize_(-1) {}
  PromMap( int nloc, const int *ElementSizeList, const PromComm &comm ) 
    :PromMap_base(nloc, sumeqs(nloc,ElementSizeList),comm),id_leq_(NULL), bsize_(-1){
    SetLidLeq( ElementSizeList );
  }
protected:
  int SetLidLeq( const int *ElementSizeList ) {
    int ii,jj,tt,s2,r2,nloc=nlocal(); 
    // determine if constant ndf
    if( nloc > 0 ) {
      for(ii=1,tt=ElementSizeList[0];ii<nloc;ii++) {
	if(ElementSizeList[ii] != tt) break;
      }
      if( ii != nloc ) s2 = tt = 0; // broke
      else s2 = tt; // high / low
    }
    else{ s2 = 0; tt = 999; }
    MPI_Allreduce(&tt, &jj, 1, MPI_INT, MPI_MIN, comm_() );
    MPI_Allreduce(&s2, &r2, 1, MPI_INT, MPI_MAX, comm_() );
    const bool constndf = (jj==r2 && jj>0); const int ndf = jj;
    if( constndf ) bsize_ = ndf;
    else{
      PetscMalloc(sizeof(int)*(nloc+1),&id_leq_); 
      for(ii=0,id_leq_[0] = 0 ; ii < nloc ; ii++ ){
	id_leq_[ii+1] = id_leq_[ii] + ElementSizeList[ii];
	assert(ElementSizeList[ii]>0);
      }
    }
    return 0;
  }

  static int sumeqs(const int nloc, const int *ElementSizeList ){
    int ii,cc;
    for(ii = cc = 0 ; ii < nloc ; ii++ )cc += ElementSizeList[ii];
    return cc;
  }
public:
  virtual ~PromMap() {
    if( id_leq_ != NULL ) PetscFree( id_leq_ ); 
  }
  //
  int nlocal() const {
    assert(this!=NULL);
    int mype = comm_.mype();
    return (proc_gnode_[mype+1] - proc_gnode_[mype]);
  }
  int nglobal() const {
    assert(this!=NULL); 
    int np = comm_.np();
    return (proc_gnode_[np]); 
  } 
  int nlocalEq() const {
    assert(this!=NULL); 
    int mype = comm_.mype();
    return (proc_geq_[mype+1] - proc_geq_[mype]);
  }
  int nglobalEq() const {
    assert(this!=NULL); 
    int np = comm_.np();
    return proc_geq_[np]; 
  }
  bool isConstNDF() const { assert(this!=NULL && bsize_ != 0); return (bsize_>0); }
  int constNDFSize() const { assert(this!=NULL && bsize_>0); return bsize_; }
  int locID_ndf(const int i)const{ 
    assert(this!=NULL); // this will fail while VBR hack is not fixed
    return (bsize_>0) ? bsize_ : id_leq_[i+1] - id_leq_[i];
  }
private:
  int *id_leq_;
  short int bsize_;
};

#undef __FUNCT__
#define __FUNCT__ "PromVector" 

//! class PromSeqVector
/*!
  - Dummy wrapper for local vector with data
*/
class PromSeqVector 
{
  friend class PromScatter;
public:
  //! dummy copy constructor for (local) salinas vectors
  PromSeqVector( int nx, double data[] ) : svec_(NULL) { 
    PetscFunctionBegin;
    VecCreateSeqWithArray( PETSC_COMM_SELF, nx, data, &svec_);
    PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
  }
  virtual ~PromSeqVector(){ VecDestroy(svec_); }
 private:
  Vec svec_;
};
//! class PromVector
/*!
  - Vec/Petra_vec
*/
class PromVector : public PromVector_base
{
  friend class PromMatrix;
  friend class PromPC;
  friend class PromZRP;
  friend class PromSmoother;
  friend class Prometheus;
  friend class pFEAP;
  friend class PromPCGaussSeidel;
  friend class PromPCKKT_ASM;
  friend class PromScatter;
public:
  //! normal constructor1
  PromVector(const PromMap &map, const int fact = 1 ) : 
    PromVector_base(map), vec_(NULL)
  {
    PetscFunctionBegin;
    VecCreateMPI(map.comm_(),map.nlocalEq()*fact,map.nglobalEq()*fact,&vec_);
    assert(vec_ != NULL);
#if defined(PETSC_USE_BOPT_g)
    double b=1.e206;
    VecSet(&b,vec_);
#endif
    PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
  }
  //! copy constructor
  PromVector( const PromVector_base * const vec_b ) : 
    PromVector_base( vec_b->map_ ), vec_(NULL) { 
    const PromVector *vec = dynamic_cast<const PromVector*>(vec_b);
    assert(vec!=NULL);
    PetscFunctionBegin;
    int ierr = VecDuplicate( vec->vec_, &vec_ );       assert(ierr==0);
    PetscStackPop; /* since function returns void cannot use PetscFunctionReturn(); */
  }
  //! destructor
  virtual ~PromVector(){ 
    //if( malled_ ) 
    VecDestroy(vec_); vec_ = NULL;
  }
  //! methods
  virtual int GetArray( double **arr ) const {
    assert(this!=NULL);
    return VecGetArray( vec_, arr );
  }
  virtual int RestoreArray( double **arr ) const {
    assert(this!=NULL);
    return VecRestoreArray( vec_, arr );
  }
  virtual double operator[](int i) const { 
    double *arr, tmp; assert(this!=NULL); 
    int ierr = VecGetArray( vec_, &arr ); CHKERRQ(ierr);
    tmp = arr[i];
    ierr = VecRestoreArray( vec_, &arr ); CHKERRQ(ierr);
    return tmp;
  }
  virtual int Set( double val ){
    assert(this!=NULL);
    return VecSet( vec_, val );
  }
  virtual int Scale( double val ) {
    assert(this!=NULL);
    return VecScale( vec_, val );
  }
  virtual int Sqrt(){
    assert(this!=NULL); // todo!
    int ierr = VecSqrt( vec_ ); CHKERRQ(ierr);
    return 0;
  }
  virtual int Dot(const PromVector_base * const b, double * const dot)const{
    assert(this!=NULL);
    const PromVector *vec = dynamic_cast<const PromVector*>(b);
    assert(vec!=NULL);
    int ierr = VecDot( vec_, vec->vec_, dot ); CHKERRQ(ierr); 
    return 0;
  }
  virtual int SetValues( const int nrow, int *rowp, const double *vals,
			 const int add_type ){
    assert(this!=NULL);
    int ierr = VecSetValues( vec_, nrow, rowp, vals, (InsertMode)add_type );
    CHKERRQ(ierr); 
    return 0;
  }
  virtual int Norm2( double * const norm ) const {
    assert(this!=NULL); PetscReal nn;
    int ierr = VecNorm( vec_, NORM_2, &nn ); *norm = (double)nn;
    CHKERRQ(ierr); 
    return 0;
  } 
  virtual int Normalize( double * const norm ) {
    assert(this!=NULL); PetscReal a;
    int ierr = VecNorm( vec_, NORM_2, &a ); CHKERRQ(ierr); 
    *norm = (double)a;
    if (a == 0.0) {
#if defined(PETSC_USE_BOPT_g)
      PetscPrintf(PETSC_COMM_WORLD,"[0]%s line %d ERROR: |x| = 0.0",__FILE__,__LINE__);
#endif
    } else {
      a = 1.0 / a;
      ierr = VecScale( vec_, a ); CHKERRQ(ierr);
    } 
    return 0;
  } 
  virtual int NormMax( double * const norm ) const {
    assert(this!=NULL); PetscReal nn;
    int ierr = VecNorm( vec_, NORM_MAX, &nn ); *norm = (double)nn;
    CHKERRQ(ierr); 
    return 0;
  } 
  virtual int Copy( const PromVector_base * const pvec ) {
    const PromVector *vec = dynamic_cast<const PromVector*>(pvec);
    assert(vec!=NULL);
    int ierr = VecCopy( vec->vec_, vec_ ); CHKERRQ(ierr); 
    return 0;
  }
  //! this = a * X + b * this
  virtual int AXPBY( double a, double b, const PromVector_base *const X ) {
    assert(this!=NULL);
    const PromVector *vec = dynamic_cast<const PromVector*>(X);
    assert(vec!=NULL);
    return VecAXPBY( vec_, a, b, vec->vec_ );
  }
  //! this <- a * this + X
  virtual int AYPX( double a, const PromVector_base *const X ) {
    const PromVector *vec = dynamic_cast<const PromVector*>(X);
    assert(vec!=NULL);
    int ierr =  VecAYPX( vec_, a, vec->vec_ ); CHKERRQ(ierr); 
    return 0;
  }
  //! this <- this + a * X
  virtual int AXPY( double a, const PromVector_base *const X ) {
    const PromVector *vec = dynamic_cast<const PromVector*>(X);
    assert(vec!=NULL);
    int ierr = VecAXPY( vec_, a, vec->vec_ ); CHKERRQ(ierr); 
    return 0;
  }
  //! this(w) = alpha x + y
  virtual int WAXPY( double alpha, const PromVector_base * const X, 
		     const PromVector_base * const Y ) {
    assert(this!=NULL);
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    return VecWAXPY( vec_, alpha, x->vec_, y->vec_ );
  }
  virtual int MAXPY(const int nv, double alpha[], const PromVector_base *const*Xarr){
    assert(this!=NULL); assert(nv<4096); assert(Xarr!=NULL);
    Vec varr[4096]; 
    for(int i=0;i<nv;i++){
      const PromVector *vv = dynamic_cast<const PromVector*>(Xarr[i]);
      assert(vv!=NULL); varr[i] = vv->vec_;
    }
    return VecMAXPY( vec_, nv, alpha, varr ); //! this = this + alpha[j] X[j] 
  }
  virtual int MDot(const int nv, const PromVector_base *const*Xarr, double arr[])const{
    assert(this!=NULL); assert(nv<4096); assert(Xarr!=NULL);
    Vec varr[4096]; 
    for(int i=0;i<nv;i++){
      const PromVector *vv = dynamic_cast<const PromVector*>(Xarr[i]);
      assert(vv!=NULL); varr[i] = vv->vec_;
    }
    PetscLogFlops(2*nv*vec_->n); 
    return VecMDot( nv, vec_, varr, arr );
  }
  virtual int Assembly(){
    assert(this!=NULL);
    int ierr = VecAssemblyBegin( vec_ ); CHKERRQ(ierr);
    return VecAssemblyEnd( vec_ );
  }
  virtual int PointwiseMult( const PromVector_base * const X,
			     const PromVector_base * const Y){
    assert(this!=NULL);  //! this <- Y * X
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    return VecPointwiseMult( vec_, x->vec_, y->vec_ );
  } 
  virtual int PointwiseDiv( const PromVector_base *const X, 
		    const PromVector_base *const Y )  {
    assert(this!=NULL);  //! this <- X / Y
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    return VecPointwiseDivide( vec_, x->vec_, y->vec_ );
  } 
  virtual int Reciprocal() {
    assert(this!=NULL);  //! this <- 1 / this
    int ierr = VecReciprocal( vec_ ); CHKERRQ(ierr); 
    return 0;
  }
  virtual int getN() const { return map_.nglobalEq(); }
  virtual int getn() const { return map_.nlocalEq(); }
  virtual int my0() const { return map_.my0Eq(); }
  //! data
  //protected:
  Vec vec_;
};

#undef __FUNCT__
#define __FUNCT__ "PromMatrix" 
//! class PromMatrix
/*!
  - PETSc PromMatrix
*/
class PromMatrix : public PromMatrix_base
{
  friend class PromPC;
  friend class PromMG;
  friend class Prometheus;
  friend class PromContext;
  friend class PromPCGaussSeidel;
  friend class PromPCKKT_ASM;
  friend class PromCRMatrix;
  friend class PromGrid;
public:
  //! matrix
  PromMatrix( const PromGrid *grid, int nA, int *pnA,int nB,int *pnB, 
	      int fact=1); 
  //! matrix
  PromMatrix( const PromMap &map, int nA, int *pnA, int nB, int *pnB, 
	      int lcols = -1, int gcols = -1, int fact=1);  
  //! common constructor
  virtual int SetUp( int nA,int *pnA,int nB,int *pnB, int lcols, int gcols, 
		     int fact=1); 
  //! Copy constructor
  PromMatrix( const PromMatrix_base* B ) : 
    PromMatrix_base(B->map_), mat_(NULL), full_mat_(NULL), array_(NULL), 
    varray_(NULL), ghostGID_BID_(NULL), ghostGID_ndf_(NULL),
    ghostBID_GID_(NULL), varr_sz_(0), malled_(1) {
    const PromMatrix *A = dynamic_cast<const PromMatrix*>(B); assert(A!=NULL);
    MatDuplicate( A->mat_, MAT_COPY_VALUES, &mat_ );
  }
  //! shell constructor
  PromMatrix( const PromMap &cmap, Mat a ) : 
    PromMatrix_base(cmap), mat_(a), full_mat_(NULL), array_(NULL), 
    varray_(NULL), ghostGID_BID_(NULL), ghostGID_ndf_(NULL), 
    ghostBID_GID_(NULL), varr_sz_(0), malled_(1)
  {
    dirty_data_ = TRUE; // user matrix is dirty w/o me adding to it
  }
  //! Transpose constructor
  virtual int Transpose( const PromMap &cmap, PromMatrix_base **out ) const {
    assert(this!=NULL); assert(0); // not used
    int ierr; Mat aa;
    ierr = MatTranspose( mat_, &aa ); CHKERRQ(ierr); 
    PromMatrix *trans = new PromMatrix( cmap, aa );
    *out = trans;
    return 0;
  }
  //! destructor
  virtual ~PromMatrix() { 
    if( malled_ ) MatDestroy(mat_); mat_ = NULL;
    if( ghostGID_BID_ != NULL ) { 
      delete ghostGID_BID_; delete ghostGID_ndf_; delete ghostBID_GID_;
    }
    if( full_mat_ != NULL )  MatDestroy( full_mat_ );
    if( varray_ != NULL )  PetscFree( varray_ );
  }
  //! matrix type methods
  virtual int ConvertToSymm();
  virtual int ConvertFromSymm();
  virtual int ConvertToSuperLU() {
    if( isSuperLU() ) return 0;
#if defined(PROM_USE_SUPERLU)
    Mat M;
    int ierr = MatConvert( mat_, MATSUPERLU_DIST, &M); CHKERRQ(ierr); 
    MatDestroy(mat_); mat_ = M;
#endif
    return 0;
  }
  virtual bool isSuperLU()const{ 
    return FALSE; 
  }
  //! methods
  virtual int GetLocalNodeRow( const int brow, const int len, int &ncols, 
			       int adjacs[], double **Aij = NULL ) const ;
private:
  int GetNodeScalarRow_PETSc( const int row, const int len, int &ncols, 
			      int Brp[], double **Bap ) const;
  int GetLocal_constBlk_NodeRow( const int brow, const int len, int &ncols, 
				 int adjacs[], double **Aij = NULL ) const ; 
  int GetSymmetricLocal_constBlk_NodeRow( const int brow, const int len, 
					  int &ncols, int adjacs[], 
					  double **Aij = NULL) const ; 
  int GetLocalVBRNodeRow( const int brow, const int len, int &ncols, 
			  int adjacs[], double **Aij = NULL ) const ;
public:
  int GetDiagonal( PromVector_base *work ) const {
    assert(this!=NULL && work != NULL);
    const PromVector *vec = dynamic_cast<const PromVector*>(work);
    assert(vec!=NULL);
    return MatGetDiagonal( mat_, vec->vec_ ); 
  } 
  //! y := A'x 
  virtual int MultTranspose( const PromVector_base * const X,
			     PromVector_base * const Y ) const {
    assert(this!=NULL);
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    return MatMultTranspose( mat_, x->vec_, y->vec_ );
  }
  //! y := x + A'w 
  virtual int MultTransposeAdd( const PromVector_base * const W,
				const PromVector_base * const X, 
				PromVector_base * const Y ) const {
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    const PromVector *w = dynamic_cast<const PromVector*>(W);assert(w!=NULL);
    return MatMultTransposeAdd( mat_, w->vec_, x->vec_, y->vec_ );
  }
  //! y := Ax 
  virtual int Mult( const PromVector_base * const X, 
		    PromVector_base * const Y ) const {
    assert(this!=NULL);
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    return MatMult( mat_, x->vec_, y->vec_ );
  }
  virtual int Mult( const PromMatrix_base *const B, 
		    PromMatrix **out ) const {
    assert(this!=NULL); Mat petsc_out;
    const PromMatrix *BB = dynamic_cast<const PromMatrix*>(B);assert(BB!=NULL);
    int ierr = MatMatMult( mat_, BB->mat_, MAT_INITIAL_MATRIX, 3.0, &petsc_out );
    CHKERRQ(ierr);
    PromMatrix *prom_out = new PromMatrix( map_, petsc_out );
    prom_out->malled_ = true;
    *out = prom_out;
    return 0;
  }
  //! y := Ax 
  virtual int KSPMult( const PromVector_base * const x, 
		       PromVector_base * const b ) const {
    return Mult(x,b);
  }
  //! Computes v3 = v2 + A * v1. 
  virtual int MultAdd( const PromVector_base * const X, 
		       const PromVector_base * const Y, 
		       PromVector_base * const W )const{
    assert(this!=NULL);
    const PromVector *x = dynamic_cast<const PromVector*>(X);assert(x!=NULL);
    const PromVector *y = dynamic_cast<const PromVector*>(Y);assert(y!=NULL);
    const PromVector *w = dynamic_cast<const PromVector*>(W);assert(w!=NULL);
    return MatMultAdd( mat_, x->vec_, y->vec_, w->vec_ );
  }
  virtual int GetOwnershipRange( int *first, int *end ) const {
    assert(this!=NULL);
    int ierr = MatGetOwnershipRange( mat_, first, end );
    return ierr;
  }
  virtual int GetRow(int row, int *ncols, int **colpp=NULL,double **arrdat=NULL)const{
    assert(this!=NULL); assert(sizeof(int) == sizeof(PetscInt));
    if ( colpp == NULL && arrdat == NULL ) {
      return MatGetRow( mat_, row, ncols, PETSC_NULL, PETSC_NULL );
    }
    else if( arrdat == NULL && colpp != NULL ) {
      return MatGetRow( mat_, row, ncols, (const PetscInt**)colpp, PETSC_NULL );
    }
    else if( colpp == NULL && arrdat != NULL ) {
      return MatGetRow( mat_, row, ncols, PETSC_NULL, (const PetscScalar**)arrdat );
    }
    else {
      return MatGetRow( mat_, row, ncols, (const PetscInt**)colpp, 
			(const PetscScalar**)arrdat );
    }
  }
  virtual int RestoreRow(int row,int*ncols,int**colpp=NULL,double**arrdat=NULL)const{
    assert(this!=NULL);
    return MatRestoreRow( mat_, row, ncols, (const PetscInt**)colpp, 
			  (const PetscScalar**)arrdat );
  }
  virtual int SetValues( const int nrow, int *pgeqi, const int ncol, int *pcol,
			 double *vals, const int add_type );
  virtual int SetValuesBlocked( const int nrow, int *pgidi, const int ncol,
				int *pcol, double *vals, const int add_type);
  virtual int ZeroRows( PromIS is, double diag = 0.0 ){
    assert(this!=NULL);
    dirty_data_ = TRUE;
    return MatZeroRowsIS( mat_, is, diag );
  }
  virtual int ZeroEntries(){
    assert(this!=NULL); int ierr;
    if( isSymmetricMatrix() ) {
      ierr = ConvertFromSymm(); CHKERRQ(ierr);
    }
    else {
      ierr = MatZeroEntries( mat_ ); CHKERRQ(ierr);
    }
    dirty_data_ = TRUE; // indefinite_ = FALSE; DRhoA_ = -2.0; -- do for coarse grids w/ helpmholtz!
    return 0;
  }
  virtual int Assembly(){
    assert(this!=NULL);
    int ierr = MatAssemblyBegin( mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    ierr = MatAssemblyEnd( mat_, MAT_FINAL_ASSEMBLY ); CHKERRQ(ierr);
    return 0;
  } 
  int FlushAssembly(){
    assert(this!=NULL);
    int ierr = MatAssemblyBegin( mat_, MAT_FLUSH_ASSEMBLY ); CHKERRQ(ierr);
    return MatAssemblyEnd( mat_, MAT_FLUSH_ASSEMBLY );
  }
  int Shift( double val ){
    assert(this!=NULL);
    return MatShift( mat_, val );
  }
  virtual int Scale(const PromVector_base *const L, 
	    const PromVector_base *const R ){
    assert(this!=NULL);
    const PromVector *l = dynamic_cast<const PromVector*>(L);
    const PromVector *r = dynamic_cast<const PromVector*>(R);
    return MatDiagonalScale( mat_,  (l != NULL) ? l->vec_ : NULL,
			     (r != NULL) ? r->vec_ : NULL );
  }
  virtual int Scale( double val ) {
    assert(this!=NULL);
    if( val == 0.0 ) return ZeroEntries();
    else return MatScale( mat_, val );
  }
  int NormFrob( double *const norm ) const {
    assert(this!=NULL); PetscReal nn;
    int ierr = MatNorm( mat_, NORM_FROBENIUS, &nn ); CHKERRQ(ierr);
    *norm = (double)nn;
    return 0;
  }  
  int NormInf( double *const norm ) const {
    assert(this!=NULL); PetscReal nn;
    if( mat_->stash.size == 1 ) {
      int ierr = MatNorm( mat_, NORM_INFINITY, &nn ); CHKERRQ(ierr);
    }
    else {
      int ierr = MatNorm( mat_, NORM_FROBENIUS, &nn ); CHKERRQ(ierr);
    }
    *norm = (double)nn/sqrt((double)mat_->N);
    return 0;
  }
  //! AXPY w/ same nnz pattern
  int AXPY( double val, const PromMatrix_base * const A ) {
    assert(this!=NULL);
    const PromMatrix *AA = dynamic_cast<const PromMatrix*>(A);assert(AA!=NULL);
    if( AA->dirty_data_ == FALSE ) {
      AA->dirty_data_ = TRUE;
      AA->SetDRho( 0.0 ); // this is collective so set here
    }
    return MatAXPY( mat_, val, AA->mat_, SAME_NONZERO_PATTERN );
  }
  int SetOption( int opt ) {
    assert(this!=NULL);
    return MatSetOption( mat_, (MatOption)opt );
  }
  virtual int getN() const { return mat_->N; }
  virtual int getn() const { return mat_->n; }
  virtual int getM() const { return map_.nglobalEq(); }
  virtual int getm() const { return map_.nlocalEq(); }
  int getMNodes()const{ return map_.nglobal(); }
  int getLocalNNZ() const {
    MatInfo info; 
    MatGetInfo( mat_, MAT_LOCAL, &info );
    return (int)(info.nz_used+.05);
  }
  int getGlobalNNZ() const {
    MatInfo info; 
    MatGetInfo( mat_, MAT_GLOBAL_SUM, &info );
    return (int)(info.nz_used+.05);
  }
  //! get MPI communicator
  MPI_Comm MPIComm() const { return map_.comm_(); }
  //! has symmetic values?
  virtual bool isSymmetricValued()const{
    assert(this != NULL && mat_ != NULL);
    if( isSymmetricMatrix() ) return TRUE;
    //else return FALSE; // diabled
    PetscTruth flg; 
    MatIsSymmetric( mat_, 0.0, &flg ); // not implemented for BAIJ!!!
    return (flg==PETSC_TRUE);
  }
  //! has symmetic storage?
  virtual bool isSymmetricMatrix()const{
    assert(this != NULL && mat_ != NULL);
    int ierr, size; MatType type;
    MPI_Comm_size( mat_->comm, &size );
    ierr = MatGetType( mat_, &type); CHKERRQ(ierr);
    if(size==1) return (strcmp(type, "seqsbaij")==0);
    else return (strcmp(type, "mpisbaij")==0);
  }
  //! create an MIS (sets nodes in 'list' to DEL or SEL)
  int CreateMIS(PromGrid *const grid, PromList<PromNode*> &list, 
		int is_del( const int, const int, bool& ) );
  int ReceiveState_NC( PromGrid *const grid,
		       PromTable<PromNode*> &gid_ghost, 
		       PromTable<NdStTAG> &gid_state,
		       const int sz, 
		       int &ndone, int &todo2, int &todo, 
		       const int rtag, const int stag, const int part,
		       MPI_Comm comm );
  int GetDeps( int gid, int &ndeps, int procs[] );
  //! get array of global IDs for ghost nodes. size=num ghosts. gid=garr[i]
  int GetGArray( int **garr, int *nghosts ) const {
    int ierr, bs, size, *out = NULL;
    MPI_Comm_size( mat_->comm, &size );
    if( size == 1 ) *nghosts = 0;
    else {
      if( isSymmetricMatrix() ) {
	SETERRQ(-1,"GetGArray for symmetric matrix not implemented ");
	//Mat_MPISBAIJ *baij = (Mat_MPISBAIJ*)mat_->data; // upper left matix
	//Mat_SeqSBAIJ *B = (Mat_SeqSBAIJ*)baij->B->data;
	//out = baij->garray;     
	//*nghosts = (out==NULL) ? 0 : B->nbs;
      }
      else { 
	ierr = MatGetBlockSize( mat_, &bs); CHKERRQ(ierr);
	if( bs == 1 ) { 
	  Mat_MPIAIJ *aij = (Mat_MPIAIJ*)mat_->data; 
	  out = aij->garray;
	  *nghosts = (out==NULL) ? 0 : aij->B->n;
	}
	else { // blocked matrix
	  Mat_MPIBAIJ *baij = (Mat_MPIBAIJ*)mat_->data; // upper left matix
	  Mat_SeqBAIJ *B = (Mat_SeqBAIJ*)baij->B->data;
	  out = baij->garray;     
	  *nghosts = (out==NULL) ? 0 : B->nbs;
	}
      }
    }
    if(garr) *garr = out;
    return 0;
  }
  int RestoreGArray( int **arr ) const { return 0; }
  bool isOK() const { assert(this!=NULL); return (mat_ != NULL); }
  virtual bool isAssembled()const{
    assert(this!=NULL); return (mat_ && mat_->assembled);
  }
  //! Print
  int Print( char *str=NULL, FILE *ifile = stderr ) const{
    int rank,idi,idj1,kk,first_row,last_row,ncols;
    const PetscInt *colpp; const PetscScalar *arrdat; 
    MatGetOwnershipRange(mat_, &first_row, &last_row);
    FILE *file; 
    if(str == NULL ) file = ifile;
    else {
      char path[128]; MPI_Comm_rank(PETSC_COMM_WORLD, &rank); 
      sprintf(path,"%s_%d", str, rank );
      file = fopen(path,"w"); assert(file!=NULL);
    }
    for( idi = first_row ; idi < last_row ; idi++ ){ 
      MatGetRow( mat_, idi, &ncols, &colpp, &arrdat );
      for ( kk = 0 ; kk < ncols ; kk++ ){
	if( arrdat[kk] == 0.0 ) continue; 
	idj1 = colpp[kk] + 1;
	fprintf(file,"%d %d %.14e\n",idi + 1, idj1, arrdat[kk]);
      }
      MatRestoreRow(mat_,idi,&ncols, &colpp,&arrdat);
    }
    //fprintf( file,"%d %d %.14e\n",last_row, last_row, 0.0 );
    fclose(file);
    return 0;
  }

  //! data
protected:
  Mat mat_;
  Mat full_mat_;
  mutable int *array_;
public:
  // vbr
  static int getLidWithLBid(const int blid, const int bs, const int nloc, 
			    const int lid_leq[],const int fact, int &);
  virtual bool isMalloced()const{ return malled_; }
  mutable double *varray_;
  PromTable<int> *ghostGID_BID_;
  PromTable<int> *ghostGID_ndf_; 
  PromTable<int> *ghostBID_GID_;
  mutable int varr_sz_;
  bool malled_;
};

//! class PromScatter
/*!
  - 
*/
class PromScatter : public PromScatter_base
{
public:
  PromScatter( const PromComm &comm, int nx, int *ix, int *iy ) : 
    PromScatter_base(comm), ISx_(0), ISy_(0), IS_i_(0), IS_r_(0), scat_(0), scat_i_(0), scat_r_(0) {
    ISCreateGeneral( PETSC_COMM_SELF, nx, ix, &ISx_ ); 
    ISCreateGeneral( comm.MPIComm(), nx, iy, &ISy_ ); 
  }
  PromScatter( const PromComm &comm, int nx, int *iy ) : 
    PromScatter_base(comm), ISx_(0), ISy_(0), IS_i_(0), IS_r_(0), scat_(0), scat_i_(0), scat_r_(0) {
    ISCreateStride( PETSC_COMM_SELF, nx, 0, 1,  &ISx_ ); 
    ISCreateGeneral( comm.MPIComm(), nx, iy, &ISy_ ); 
  }
  virtual ~PromScatter(){
    ISDestroy(ISx_); ISDestroy(ISy_); 
    if(scat_) VecScatterDestroy(scat_);
    if(scat_i_) VecScatterDestroy(scat_i_);
    if(scat_r_) VecScatterDestroy(scat_r_);
    if(IS_i_) ISDestroy(IS_i_);
    if(IS_r_) ISDestroy(IS_r_);
  }
  //!
  virtual int Scatter( const PromSeqVector *X, PromVector_base *Y );
  virtual int Gather( PromSeqVector *X, const PromVector_base *Y);
  virtual int Gather_i( PromSeqVector *X, const PromVector_base *Y);
  virtual int Gather_r( PromSeqVector *X, const PromVector_base *Y);
  //! access methods
  virtual int getYIDs( int **ids )const{ ISGetIndices( ISy_, ids ); return 0;}
  virtual int restoreYIDs( int **ids )const{ ISRestoreIndices( ISy_, ids ); return 0;}
  virtual int getYID_size()const{ int s; ISGetLocalSize( ISy_, &s ); return s;}
private:
  int scatter_private( ScatterMode mode, Vec xx, Vec yy ){ 
    int ierr;
    if( scat_ == 0 ) {
      ierr = VecScatterCreate( xx, ISx_, yy, ISy_, &scat_); CHKERRQ(ierr);
    }
    if( mode == SCATTER_REVERSE ) {
      ierr = VecScatterBegin(yy, xx, INSERT_VALUES, mode, scat_); CHKERRQ(ierr);
      ierr = VecScatterEnd(yy, xx, INSERT_VALUES, mode, scat_); CHKERRQ(ierr);//
    }
    else {
      ierr = VecScatterBegin(xx, yy, INSERT_VALUES,mode,scat_); CHKERRQ(ierr);
      ierr = VecScatterEnd(xx, yy, INSERT_VALUES,mode,scat_); CHKERRQ(ierr);//
    }
    return 0;
  }
  //
  IS ISx_, ISy_, IS_i_, IS_r_;
  VecScatter scat_, scat_i_, scat_r_;
};

#undef __FUNCT__
#define __FUNCT__ "PromPC" 
//! class PromPC
/*!
  - Wrapper for PETSc PC, all low level PCs (ie, operator supplied PCs)
*/
class PromPC : public PromPC_base
{
  friend class PromPCGaussSeidel;
  friend class PromMG;
  friend class PromPCKKT_ASM;
  friend class PromPCKKT_Seg;
  friend class PromPCKKT_Shell;
public :
  //! constructor
  PromPC(const PromOptions &opt, PromPerfMonitor &perf,
	 const MPI_Comm comm) : PromPC_base(opt,perf),
				pc_(NULL), work_(NULL), rwork_(NULL) {
    PetscFunctionBegin;
    PCCreate( comm, &pc_ );
    PetscStackPop;
  }
  //! destructor
  virtual ~PromPC(){
    if( pc_ != NULL ) PCDestroy(pc_); 
    if( work_ != NULL ){ VecDestroy(work_); VecDestroy(rwork_); }
  }
  //! SetType, sets type and vector to clone if need be
  virtual int SetType( PromPCType type, const PromVector_base * const vB );
  //! getType
  virtual PromPCType getType() const {
    assert( this != NULL );   PetscTruth flag;
    // need to translate
    PetscTypeCompare((PetscObject)pc_, PCILU, &flag);
    if(flag) return PROMPCILU;
    
    PetscTypeCompare((PetscObject)pc_, PCLU, &flag);
    if(flag) return PROMPCLU;
    
    PetscTypeCompare((PetscObject)pc_, PCJACOBI, &flag);
    if(flag) return PROMPCJACOBI;
    
    PetscTypeCompare((PetscObject)pc_, PCASM, &flag);
    if(flag) return PROMPCASM;

    PetscTypeCompare((PetscObject)pc_, PCNONE, &flag);
    if(flag) return PROMPCNONE;
    
    PetscPrintf(PETSC_COMM_WORLD,"[?]%s line %d ERROR: bad pc type???\n",
		__FILE__,__LINE__);
    exit(1);
    return PROMPCILU;
  }
  virtual int SetAbstractOperator_private(const PromMatrix_base * const Ab){
    const PromMatrix *AA=dynamic_cast<const PromMatrix*>(Ab);assert(AA!=NULL);
    assert(pc_->setupcalled != 2); // should be called before setup!
    pc_->pmat = AA->mat_; // hard set of matrix
    pc_->mat = AA->mat_; assert(AA->mat_ != NULL);
    return 0;
  }
protected:
  //! setup - matrix factors
  int SetUp_private(){
#if defined(PROM_TRACE_PC_SETUP)
    PetscPrintf(pc_->comm,
		"\t\t[0]PromPC::SetUp_private this=%p setupcalled=%d\n",
		this,pc_->setupcalled);
#endif
    assert(this!=NULL);      int ierr; // new data, refactor
    ierr = PCSetUp( pc_ );                      CHKERRQ(ierr);
    ierr = PCSetUpOnBlocks( pc_ );              CHKERRQ(ierr);
    assert(pc_->setupcalled == 2);
    return 0;
  }
  //! unsets "setupcalled" flag 
  virtual int SetOperator_private( const PromMatrix_base * const A ){
    assert(this!=NULL); 
    if( pc_ == NULL ) SETERRQ(-1,"pc_ == NULL");
    const PromMatrix *AA=dynamic_cast<const PromMatrix*>(A);assert(AA!=NULL);
    // ierr = PCSetOperators( pc_, AA->mat_, AA->mat_, SAME_NONZERO_PATTERN );
    pc_->mat  = AA->mat_; // petsc does a crazy reset in PCSetOperators
    pc_->pmat = AA->mat_;
    pc_->flag = SAME_NONZERO_PATTERN;
    if( pc_->setupcalled == 2 ) pc_->setupcalled = 1; // ** soft reset

    return 0;
  }
  //! unsets "setupcalled" flag 
  virtual int SetNewOperator_private( const PromMatrix_base * const A ){
    assert(this!=NULL); 
    if( pc_ == NULL ) SETERRQ(-1,"pc_ == NULL");
    const PromMatrix *AA=dynamic_cast<const PromMatrix*>(A);assert(AA!=NULL);
    // ierr = PCSetOperators( pc_, AA->mat_, AA->mat_, SAME_NONZERO_PATTERN );
    pc_->mat  = AA->mat_; // petsc does a crazy reset in PCSetOperators
    pc_->pmat = AA->mat_;
    pc_->flag = DIFFERENT_NONZERO_PATTERN;
    if( pc_->setupcalled == 2 ) pc_->setupcalled = 1; // ** soft reset

    return 0;
  }
public:
  //! construct a Additive Scwartz Methods PC
  int CreateASM( const int nblocks, PromIS *blockISs );

  //! getPCString
  virtual const char * getPCString() const {
    assert(this!=NULL); assert(pc_!=NULL);
    PetscFunctionBegin;
    PCType pctype;
    PCGetType( pc_, &pctype );
    if( kkt_ == NULL ){ 
      PetscTruth isasm;   
      PetscTypeCompare((PetscObject)pc_,PCASM,&isasm); 
      if( isasm ){
	int nlocal, first; KSP *psles; PC subpc;
	int ierr = PCASMGetSubKSP( pc_, &nlocal, &first, &psles );
	if(ierr) return  "ERROR-1";
	KSPGetPC( psles[0], &subpc ); PCType subpctype;
	PCGetType( subpc, &subpctype );
	PetscTypeCompare((PetscObject)pc_,PCILU,&isasm); 
	if(isasm) pctype = subpctype;
      }
    }
    PetscFunctionReturn(pctype);
  }
  //! getSubPCString
  virtual const char * getSubPCString() const {
    assert(this!=NULL); assert(pc_!=NULL);
    if( kkt_ != NULL ) return kkt_->getPCString(); 
    PetscTruth isasm;   PCType subpctype;
    PetscTypeCompare((PetscObject)pc_,PCASM,&isasm); 
    if( isasm ){
      int nlocal, first; KSP *psles; PC subpc;
      int ierr = PCASMGetSubKSP( pc_, &nlocal, &first, &psles );
      if(ierr) return "ERROR-2";
      KSPGetPC( psles[0], &subpc );
      PCGetType( subpc, &subpctype );
      return subpctype;
    }
    else return PromPCStrings[PROMPCNONE]; 
  }
  //! getNumBlocks
  virtual int getNumBlocks() const {
    assert(this!=NULL); 
    PetscFunctionBegin;
    if( pc_ == NULL ) SETERRQ(-1,"pc_ == NULL");
    PetscTruth isasm; int ierr;
    ierr = PetscTypeCompare( (PetscObject)pc_, PCASM, &isasm ); CHKERRQ(ierr);
    if( isasm ) {
      int nlocal, first; KSP *psles; 
      ierr = PCASMGetSubKSP( pc_, &nlocal, &first, &psles );  CHKERRQ(ierr);
      PetscFunctionReturn(nlocal);     
    }
    else if( kkt_ != NULL ) PetscFunctionReturn(kkt_->getNumBlocks()); 
    PetscFunctionReturn(0);
  }
  //! PC  Apply 
  virtual int Apply_private(const PromVector_base * const bbB, 
			    PromVector_base * const xxB,const bool zerox=TRUE);
  //! data
protected:
  PC  pc_;
  Vec work_;
  Vec rwork_;
}; // PromPC

// want access to PETSc's block data -- hack of all hacks!!!
typedef struct { 
  int        n,n_local,n_local_true;
  int        is_flg;            /* flg set to 1 if the IS created in pcsetup */
  int        overlap;           /* overlap requested by user */
  KSP       *ksp;             /* linear solvers for each block */
  VecScatter *scat;             /* mapping to subregion */
  Vec        *x,*y;
  IS         *is;               /* index set that defines each subdomain */
  Mat        *mat,*pmat;        /* mat is not currently used */
  PCASMType  type;        /* use reduced interpolation, restriction or both */
  int        same_local_solves; /* flag indicating whether all local solvers are same */
  int        inplace;    /* indicates that the sub-matrices are deleted after 
                                PCSetUpOnBlocks() is done. Similar to inplace 
                                1  factor_ization in the case of LU and ILU */
}PETSc_ASM;

class PromGSNode;
int PromDoSchwarzBlocks(Mat AA, const int nb,const int blocks[],PETSc_ASM *osm,
			const PromVector *const bb, PromVector *const xx,
			const int ndf, const int my0,
			const int my0Eq, const bool havgh, const bool pre_s, 
			const PromTable<PromGSNode*> *gid_ghost, 
			const double ghostx[], int &numops );

#endif // __PROM_PETSC_H__
